@@ -29,15 +29,15 @@ mod memory_storage_tests {
     #[tokio::test]
     #[serial]
     async fn test_v2_in_memory_db_with_caching() {
-        let mut db = crate::storage::memory::AsyncInMemoryDbWithCache::new();
-        crate::storage::tests::run_test_cases_for_storage_impl(&mut db).await;
+        let db = crate::storage::memory::AsyncInMemoryDbWithCache::new();
+        crate::storage::tests::run_test_cases_for_storage_impl(&db).await;
     }
 
     #[tokio::test]
     #[serial]
     async fn test_v2_in_memory_db() {
-        let mut db = AsyncInMemoryDatabase::new();
-        crate::storage::tests::run_test_cases_for_storage_impl(&mut db).await;
+        let db = AsyncInMemoryDatabase::new();
+        crate::storage::tests::run_test_cases_for_storage_impl(&db).await;
     }
 }
 
@@ -45,11 +45,13 @@ mod memory_storage_tests {
 /// Run the storage-layer test suite for a given storage implementation.
 /// This is public because it can be used by other implemented storage layers
 /// for consistency checks (e.g. mysql, memcached, etc)
-pub async fn run_test_cases_for_storage_impl<S: Storage + Sync + Send>(db: &mut S) {
+pub async fn run_test_cases_for_storage_impl<S: Storage + Sync + Send>(db: &S) {
+    crate::test_utils::init_logger(log::Level::Info);
     test_get_and_set_item(db).await;
     test_user_data(db).await;
     test_transactions(db).await;
     test_batch_get_items(db).await;
+    test_tombstoning_data(db).await.unwrap();
 }
 
 // *** New Test Helper Functions *** //
@@ -64,7 +66,7 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
     assert_eq!(Ok(()), set_result);
 
     let get_result = storage
-        .get::<Azks>(crate::append_only_zks::DEFAULT_AZKS_KEY)
+        .get::<Azks>(&crate::append_only_zks::DEFAULT_AZKS_KEY)
         .await;
     if let Ok(DbRecord::Azks(got_azks)) = get_result {
         assert_eq!(got_azks.latest_epoch, azks.latest_epoch);
@@ -76,17 +78,17 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
     // === HistoryTreeNode storage === //
 
     let node = HistoryTreeNode {
-        label: NodeLabel::new(13, 4),
+        label: NodeLabel::new(byte_arr_from_u64(13), 4),
         birth_epoch: 123,
         last_epoch: 234,
-        parent: NodeLabel::new(1, 1),
+        parent: NodeLabel::new(byte_arr_from_u64(1), 1),
         node_type: NodeType::Leaf,
     };
     let mut node2 = node.clone();
-    node2.label = NodeLabel::new(16, 4);
+    node2.label = NodeLabel::new(byte_arr_from_u64(16), 4);
 
-    let key = NodeKey(NodeLabel::new(13, 4));
-    let key2 = NodeKey(NodeLabel::new(16, 4));
+    let key = NodeKey(NodeLabel::new(byte_arr_from_u64(13), 4));
+    let key2 = NodeKey(NodeLabel::new(byte_arr_from_u64(16), 4));
 
     let set_result = storage.set(DbRecord::HistoryTreeNode(node.clone())).await;
     assert_eq!(Ok(()), set_result);
@@ -94,7 +96,7 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
     let set_result = storage.set(DbRecord::HistoryTreeNode(node2.clone())).await;
     assert_eq!(Ok(()), set_result);
 
-    let get_result = storage.get::<HistoryTreeNode>(key).await;
+    let get_result = storage.get::<HistoryTreeNode>(&key).await;
     if let Ok(DbRecord::HistoryTreeNode(got_node)) = get_result {
         assert_eq!(got_node.label, node.label);
         assert_eq!(got_node.parent, node.parent);
@@ -105,15 +107,15 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
         panic!("Failed to retrieve History Tree Node");
     }
 
-    let get_result = storage.get::<HistoryTreeNode>(key2).await;
+    let get_result = storage.get::<HistoryTreeNode>(&key2).await;
     if let Err(err) = get_result {
         panic!("Failed to retrieve history tree node (2) {:?}", err)
     }
 
     // === HistoryNodeState storage === //
-    let key = NodeStateKey(NodeLabel::new(1, 1), 1);
+    let key = NodeStateKey(NodeLabel::new(byte_arr_from_u64(1), 1), 1);
     let node_state = HistoryNodeState {
-        value: vec![],
+        value: [0u8; 32],
         child_states: [None, None],
         key,
     };
@@ -122,7 +124,7 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
         .await;
     assert_eq!(Ok(()), set_result);
 
-    let get_result = storage.get::<HistoryNodeState>(key).await;
+    let get_result = storage.get::<HistoryNodeState>(&key).await;
     if let Ok(DbRecord::HistoryNodeState(got_state)) = get_result {
         assert_eq!(got_state.value, node_state.value);
         assert_eq!(got_state.child_states, node_state.child_states);
@@ -132,18 +134,18 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
     }
 
     // === ValueState storage === //
-    let key = ValueStateKey("test".to_string(), 1);
+    let key = ValueStateKey("test".as_bytes().to_vec(), 1);
     let value = ValueState {
-        username: AkdLabel("test".to_string()),
+        username: AkdLabel::from_utf8_str("test"),
         epoch: 1,
-        label: NodeLabel::new(1, 1),
+        label: NodeLabel::new(byte_arr_from_u64(1), 1),
         version: 1,
-        plaintext_val: AkdValue("abc123".to_string()),
+        plaintext_val: AkdValue::from_utf8_str("abc123"),
     };
     let set_result = storage.set(DbRecord::ValueState(value.clone())).await;
     assert_eq!(Ok(()), set_result);
 
-    let get_result = storage.get::<ValueState>(key).await;
+    let get_result = storage.get::<ValueState>(&key).await;
     if let Ok(DbRecord::ValueState(got_state)) = get_result {
         assert_eq!(got_state.username, value.username);
         assert_eq!(got_state.epoch, value.epoch);
@@ -156,15 +158,14 @@ async fn test_get_and_set_item<Ns: Storage>(storage: &Ns) {
 }
 
 async fn test_batch_get_items<Ns: Storage>(storage: &Ns) {
-    let mut rand_users: Vec<String> = vec![];
+    let mut rand_users: Vec<Vec<u8>> = vec![];
     for _ in 0..20 {
-        rand_users.push(
-            thread_rng()
-                .sample_iter(&Alphanumeric)
-                .take(30)
-                .map(char::from)
-                .collect(),
-        );
+        let str: String = thread_rng()
+            .sample_iter(&Alphanumeric)
+            .take(30)
+            .map(char::from)
+            .collect();
+        rand_users.push(str.as_bytes().to_vec());
     }
 
     let mut data = Vec::new();
@@ -176,7 +177,7 @@ async fn test_batch_get_items<Ns: Storage>(storage: &Ns) {
                 plaintext_val: AkdValue(value.clone()),
                 version: epoch,
                 label: NodeLabel {
-                    val: 1u64,
+                    val: byte_arr_from_u64(1),
                     len: 1u32,
                 },
                 epoch,
@@ -191,7 +192,7 @@ async fn test_batch_get_items<Ns: Storage>(storage: &Ns) {
     let toc: Duration = Instant::now() - tic;
     println!("Storage batch op: {} ms", toc.as_millis());
     let got = storage
-        .get::<ValueState>(ValueStateKey(rand_users[0].clone(), 10))
+        .get::<ValueState>(&ValueStateKey(rand_users[0].clone(), 10))
         .await;
     if got.is_err() {
         panic!("Failed to retrieve a user after batch insert");
@@ -201,7 +202,7 @@ async fn test_batch_get_items<Ns: Storage>(storage: &Ns) {
         .iter()
         .map(|user| ValueStateKey(user.clone(), 1))
         .collect();
-    let got_all = storage.batch_get::<ValueState>(keys).await;
+    let got_all = storage.batch_get::<ValueState>(&keys).await;
     match got_all {
         Err(_) => panic!("Failed to retrieve batch of user at specific epochs"),
         Ok(lst) if lst.len() != rand_users.len() => {
@@ -319,16 +320,15 @@ async fn test_batch_get_items<Ns: Storage>(storage: &Ns) {
     }
 }
 
-async fn test_transactions<S: Storage + Sync + Send>(storage: &mut S) {
-    let mut rand_users: Vec<String> = vec![];
+async fn test_transactions<S: Storage + Sync + Send>(storage: &S) {
+    let mut rand_users: Vec<Vec<u8>> = vec![];
     for _ in 0..20 {
-        rand_users.push(
-            thread_rng()
-                .sample_iter(&Alphanumeric)
-                .take(30)
-                .map(char::from)
-                .collect(),
-        );
+        let str: String = thread_rng()
+            .sample_iter(&Alphanumeric)
+            .take(30)
+            .map(char::from)
+            .collect();
+        rand_users.push(str.as_bytes().to_vec());
     }
 
     let mut data = Vec::new();
@@ -340,7 +340,7 @@ async fn test_transactions<S: Storage + Sync + Send>(storage: &mut S) {
                 plaintext_val: AkdValue(value.clone()),
                 version: 1u64,
                 label: NodeLabel {
-                    val: 1u64,
+                    val: byte_arr_from_u64(1),
                     len: 1u32,
                 },
                 epoch,
@@ -369,7 +369,7 @@ async fn test_transactions<S: Storage + Sync + Send>(storage: &mut S) {
     let toc: Duration = Instant::now() - tic;
     println!("Storage batch op: {} ms", toc.as_millis());
     let got = storage
-        .get::<ValueState>(ValueStateKey(rand_users[0].clone(), 10))
+        .get::<ValueState>(&ValueStateKey(rand_users[0].clone(), 10))
         .await;
     if got.is_err() {
         panic!("Failed to retrieve a user after batch insert");
@@ -383,7 +383,7 @@ async fn test_transactions<S: Storage + Sync + Send>(storage: &mut S) {
     println!("Transactional storage batch op: {} ms", toc.as_millis());
 
     let got = storage
-        .get::<ValueState>(ValueStateKey(rand_users[0].clone(), 10 + 10000))
+        .get::<ValueState>(&ValueStateKey(rand_users[0].clone(), 10 + 10000))
         .await;
     if got.is_err() {
         panic!("Failed to retrieve a user after batch insert");
@@ -391,28 +391,32 @@ async fn test_transactions<S: Storage + Sync + Send>(storage: &mut S) {
 }
 
 async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
-    let rand_user: String = thread_rng()
+    let rand_user = thread_rng()
         .sample_iter(&Alphanumeric)
         .take(30)
         .map(char::from)
-        .collect();
-    let rand_value: String = thread_rng()
+        .collect::<String>()
+        .as_bytes()
+        .to_vec();
+    let rand_value = thread_rng()
         .sample_iter(&Alphanumeric)
         .take(1028)
         .map(char::from)
-        .collect();
+        .collect::<String>()
+        .as_bytes()
+        .to_vec();
     let mut sample_state = ValueState {
         plaintext_val: AkdValue(rand_value.clone()),
         version: 1u64,
         label: NodeLabel {
-            val: 1u64,
+            val: byte_arr_from_u64(1),
             len: 1u32,
         },
         epoch: 1u64,
         username: AkdLabel(rand_user),
     };
     let mut sample_state_2 = sample_state.clone();
-    sample_state_2.username = AkdLabel("test_user".to_string());
+    sample_state_2.username = AkdLabel::from_utf8_str("test_user");
 
     let result = storage
         .set(DbRecord::ValueState(sample_state.clone()))
@@ -480,7 +484,7 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
         Ok(ValueState {
             epoch: 123,
             version: 2,
-            label: NodeLabel::new(1, 1),
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
             plaintext_val: AkdValue(rand_value.clone()),
             username: sample_state.username.clone(),
         }),
@@ -488,14 +492,14 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
     );
 
     let specifc_result = storage
-        .get::<ValueState>(ValueStateKey(sample_state.username.0.clone(), 123))
+        .get::<ValueState>(&ValueStateKey(sample_state.username.to_vec(), 123))
         .await;
     if let Ok(DbRecord::ValueState(state)) = specifc_result {
         assert_eq!(
             ValueState {
                 epoch: 123,
                 version: 2,
-                label: NodeLabel::new(1, 1),
+                label: NodeLabel::new(byte_arr_from_u64(1), 1),
                 plaintext_val: AkdValue(rand_value.clone()),
                 username: sample_state.username.clone(),
             },
@@ -511,7 +515,7 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
             ValueStateRetrievalFlag::SpecificVersion(100),
         )
         .await;
-    assert!(matches!(missing_result, Err(StorageError::GetData(_)),));
+    assert!(matches!(missing_result, Err(StorageError::NotFound(_)),));
 
     let specific_result = storage
         .get_user_state(
@@ -523,7 +527,7 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
         Ok(ValueState {
             epoch: 123,
             version: 2,
-            label: NodeLabel::new(1, 1),
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
             plaintext_val: AkdValue(rand_value.clone()),
             username: sample_state.username.clone(),
         }),
@@ -537,7 +541,7 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
         Ok(ValueState {
             epoch: 1,
             version: 1,
-            label: NodeLabel::new(1, 1),
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
             plaintext_val: AkdValue(rand_value.clone()),
             username: sample_state.username.clone(),
         }),
@@ -551,7 +555,7 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
         Ok(ValueState {
             epoch: 456,
             version: 3,
-            label: NodeLabel::new(1, 1),
+            label: NodeLabel::new(byte_arr_from_u64(1), 1),
             plaintext_val: AkdValue(rand_value.clone()),
             username: sample_state.username.clone(),
         }),
@@ -582,3 +586,93 @@ async fn test_user_data<S: Storage + Sync + Send>(storage: &S) {
     let data = storage.get_user_data(&sample_state_2.username).await;
     assert_eq!(4, data.unwrap().states.len());
 }
+
+async fn test_tombstoning_data<S: Storage + Sync + Send>(
+    storage: &S,
+) -> Result<(), crate::errors::AkdError> {
+    let rand_user = thread_rng()
+        .sample_iter(&Alphanumeric)
+        .take(30)
+        .map(char::from)
+        .collect::<String>()
+        .as_bytes()
+        .to_vec();
+    let rand_value = rand_user.clone();
+
+    let mut sample_state = ValueState {
+        plaintext_val: AkdValue(rand_value.clone()),
+        version: 1u64,
+        label: NodeLabel {
+            val: byte_arr_from_u64(1),
+            len: 1u32,
+        },
+        epoch: 1u64,
+        username: AkdLabel(rand_user.clone()),
+    };
+    let mut sample_state2 = sample_state.clone();
+    sample_state2.username = AkdLabel::from_utf8_str("tombstone_test_user");
+
+    // Load up a bunch of data into the storage layer
+    for i in 0..5 {
+        sample_state.version = i;
+        sample_state.epoch = i;
+        sample_state2.version = i;
+        sample_state2.epoch = i;
+
+        assert_eq!(
+            Ok(()),
+            storage
+                .set(DbRecord::ValueState(sample_state.clone()))
+                .await
+        );
+        assert_eq!(
+            Ok(()),
+            storage
+                .set(DbRecord::ValueState(sample_state2.clone()))
+                .await
+        );
+    }
+
+    let data = storage.get_user_data(&sample_state.username).await.unwrap();
+    assert_eq!(5, data.states.len());
+    let data = storage
+        .get_user_data(&sample_state2.username)
+        .await
+        .unwrap();
+    assert_eq!(5, data.states.len());
+
+    let keys_to_tombstone = [
+        ValueStateKey("tombstone_test_user".as_bytes().to_vec(), 0u64),
+        ValueStateKey("tombstone_test_user".as_bytes().to_vec(), 1u64),
+        ValueStateKey(sample_state.username.to_vec(), 0u64),
+        ValueStateKey(sample_state.username.to_vec(), 1u64),
+        ValueStateKey(sample_state.username.to_vec(), 2u64),
+    ];
+
+    // tombstone the given states
+    storage.tombstone_value_states(&keys_to_tombstone).await?;
+
+    for label in [
+        AkdLabel::from_utf8_str("tombstone_test_user"),
+        AkdLabel(rand_user),
+    ] {
+        for version in 0..5 {
+            let key = ValueStateKey(label.to_vec(), version);
+            let got = storage.get::<ValueState>(&key).await?;
+
+            if let DbRecord::ValueState(value_state) = got {
+                assert_eq!(version, value_state.epoch);
+                if keys_to_tombstone.contains(&key) {
+                    // should be a tombstone
+                    assert_eq!(crate::TOMBSTONE.to_vec(), value_state.plaintext_val.0);
+                } else {
+                    // should NOT be a tombstone
+                    assert_ne!(crate::TOMBSTONE.to_vec(), value_state.plaintext_val.0);
+                }
+            } else {
+                panic!("Unable to retrieve value state {:?}", key);
+            }
+        }
+    }
+    Ok(())
+}
